from yta_editor_nodes_cpu.processor import _NodeProcessorCPU
from yta_programming.decorators.requires_dependency import requires_dependency
from abc import abstractmethod

import numpy as np


class _VideoNodeProcessorCPU(_NodeProcessorCPU):
    """
    *Abstract class*

    *Singleton class*

    *For internal use only*

    Class to represent a node processor that uses CPU
    to transform the input but it is meant to video
    processing, so it implements a 'time' parameter to
    manipulate the frames according to that time
    moment.

    This class must be implemented by any processor
    that uses CPU to modify an input.
    """
    
    @abstractmethod
    def process(
        self,
        # TODO: What about the type (?)
        input: np.ndarray,
        t: float,
        **kwargs
    # TODO: What about the output type (?)
    ) -> np.ndarray:
        """
        Process the provided `input` and transform it by
        using the code that is defined here according to
        the `t` time moment provided.
        """
        # TODO: Specific attributes can be received as
        # **kwargs to modify the specific process
        pass

# Specific implementations here below:
class WavingFramesVideoNodeProcessorCPU(_VideoNodeProcessorCPU):
    """
    Just an example of a specific class that is a node
    processor that uses CPU.

    TODO: Untested and needing 'opencv-python' as 'cv2'
    """

    def __init__(
        self,
        amplitude: float = 0.05,
        frequency: float = 10.0,
        speed: float = 2.0
    ):
        """
        time: tiempo actual (float)
        amplitude: desplazamiento vertical máximo (en píxeles o en [0,1] si normalizado)
        frequency: frecuencia de la onda (ciclos por ancho)
        speed: velocidad de movimiento (en radianes / segundo)
        """
        self._amplitude: float = amplitude
        self._frequency: float = frequency
        self._speed: float = speed

    @requires_dependency('cv2', 'yta_editor_nodes_cpu', 'opencv-python')
    def process(
        self,
        # TODO: What about the type (?)
        input: np.ndarray,
        t: float
    # TODO: What about the output type (?)
    ) -> np.ndarray:
        """
        Process the provided 'input' and transform it by
        using the code that is defined here.
        """
        import numpy as np
        import cv2  # opcional, solo para redimensionar/interpolar

        h, w, _ = input.shape

        # UV coordinates [0, 1]
        x = np.linspace(0, 1, w, dtype = np.float32)
        y = np.linspace(0, 1, h, dtype = np.float32)
        xv, yv = np.meshgrid(x, y)

        wave = np.sin(xv * self._frequency + t * self._speed) * self._amplitude

        # UV coordinates displacement
        yv_new = yv + wave

        # Imitate wrap/clamp of OpenGL (comment if REPEAT)
        yv_new = np.clip(yv_new, 0.0, 1.0)

        # UV coordinates to pixels
        map_x = (xv * (w - 1)).astype(np.float32)
        map_y = (yv_new * (h - 1)).astype(np.float32)

        return cv2.remap(
            input,
            map_x,
            map_y,
            interpolation = cv2.INTER_LINEAR,
            borderMode = cv2.BORDER_REFLECT
        )


"""
We don't name it 'blend' instead of 'process' because
the GPU classes have the 'process' name due to the
inheritance from OpenGL classes, and we want to avoid
misunderstandings.
"""
# TODO: Should we keep this functionality (?)
# from yta_video_opengl.utils import texture_to_frame
from yta_editor_nodes_cpu.processor.blender.utils import _validate_inputs_and_mix_weights, _ensure_uint8
from yta_programming.singleton import SingletonABCMeta
from yta_validation.parameter import ParameterValidator
from yta_validation import PythonValidator
from abc import abstractmethod
from typing import Union

import numpy as np


class _BlenderCPU(metaclass = SingletonABCMeta):
    """
    *Singleton class*

    *For internal use only*

    Class to represent a blender that uses CPU to
    blend the inputs.

    This class must be implemented by any blender
    that uses CPU to blend inputs.
    """

    @abstractmethod
    def _blend(
        self,
        base_input: np.ndarray,
        overlay_input: np.ndarray,
        **kwargs
    # TODO: Is this the expected type (?)
    # TODO: What about OpenGL textures (?)
    ) -> np.ndarray:
        """
        *For internal use only*

        *This method must be overwritten by the specific
        classes*

        The internal process to blend and mix the provided
        `base_input` and `overlay_input`.

        This method should not force uint8 nor [0, 255]
        range by itself as it would be done in the 'blend'
        main method.
        """
        pass

    def process(
        self,
        base_input: Union[np.ndarray, 'moderngl.Texture'],
        overlay_input: Union[np.ndarray, 'moderngl.Texture'],
        mix_weight: float = 1.0,
        dtype: Union[np.dtype, None] = None,
        **kwargs
    # TODO: Is this the expected type (?)
    # TODO: What about OpenGL textures (?)
    ) -> np.ndarray:
        """
        Blend the `base_input` provided with the also given
        `overlay_input`, using the `mix_weight` to obtain
        a single output as the result.

        The inputs should be received in uint8 format in
        the [0, 255] range, then processed with the specific
        processing method, and then turned back again to the
        uint8 format and [0, 255] range.
        """
        ParameterValidator.validate_mandatory_instance_of('base_input', base_input, [np.ndarray, 'moderngl.Texture'])
        ParameterValidator.validate_mandatory_instance_of('overlay_input', overlay_input, [np.ndarray, 'moderngl.Texture'])
        ParameterValidator.validate_mandatory_number_between('mix_weight', mix_weight, 0.0, 1.0)

        # TODO: We should always receive the inputs as
        # uint8 in [0, 255] range

        # TODO: We have to make sure the sizes are the
        # same, or force them or do something...
        _validate_inputs_and_mix_weights(
            inputs = [base_input, overlay_input],
            mix_weights = [mix_weight]
        )

        # TODO: Should we keep this functionality (?)
        # Transform to numpy arrays if textures received
        # base_input = (
        #     texture_to_frame(base_input, do_include_alpha = True)
        #     if PythonValidator.is_instance_of(base_input, 'moderngl.Texture') else
        #     base_input
        # )

        # overlay_input = (
        #     texture_to_frame(overlay_input, do_include_alpha = True)
        #     if PythonValidator.is_instance_of(overlay_input, 'moderngl.Texture') else
        #     overlay_input
        # )

        dtype = (
            base_input.dtype
            if dtype is None else
            dtype
        )

        # TODO: Size of 'base_input' and 'overlay_input' must
        # be the same, and the 'base_input' should be the size
        if mix_weight == 0:
            # We want to affect 0% to the base, so we return the
            # base directly instead of calculating, but... Y_Y
            return base_input

        # 1. Blend with the specific internal process
        blended = self._blend(
            base_input = base_input,
            overlay_input = overlay_input,
            **kwargs
        )

        # 2. Apply mix_weight to determine the percentage to affect
        blended = (
            base_input * (1 - mix_weight) + blended * mix_weight
            if mix_weight < 1.0 else
            blended
        )

        # Always turn back to uint8 in [0, 255] range
        blended = _ensure_uint8(blended)

        return blended.astype(dtype)

    def process_multiple_inputs(
        self,
        inputs: list[Union[np.ndarray, 'moderngl.Texture']],
        mix_weights: Union[list[float], float] = 1.0,
        dtype: Union[np.dtype, None] = None,
        **kwargs
    ) -> np.ndarray:
        """
        Blend all the `inputs` provided, one after another,
        applying the `mix_weight` provided, and forcing the
        result to the `dtype` if provided.

        The `mix_weight` can be a single float value, that 
        will be used for all the mixings, or a list of as
        many float values as `inputs` received, to be 
        applied individually to each mixing.
        """
        _validate_inputs_and_mix_weights(
            inputs = inputs,
            mix_weights = mix_weights
        )
        
        # We process all the 'inputs' as 'base' and 'overlay'
        # and accumulate the result
        dtype = (
            inputs[0].dtype
            if dtype is None else
            dtype
        )

        # Use the first one as the base
        base = inputs[0]

        # TODO: How do we handle the additional parameters that
        # could be an array? Maybe if it is an array, check that
        # the number of elements is the same as the number of
        # inputs, and if a single value just use it...

        for i in range(1, len(inputs)):
            overlay = inputs[i]
            mix_weight = mix_weights[i]

            # TODO: If we make a lot of different operations and
            # force the [0, 255] range in each of them, we could
            # have an accumulated error... Think if we need to
            # refactor this
            base = self.process(
                base_input = base,
                overlay_input = overlay,
                mix_weight = mix_weight
            )

        # Result is always forced to uint8 and [0, 255] when
        # processed by pairs, so we don't need  to do it
        # TODO: Maybe we can avoid 'dtype' parameter as it
        # is not useful at all...
        return base.astype(dtype)
    
# Specific implementations below
class MixBlenderCPU(_BlenderCPU):
    """
    Blend the second input with the first one
    applying the float factor (between 0.0 and
    1.0) that is passed as parameter.
    """

    def _blend(
        self,
        base_input: np.ndarray,
        overlay_input: np.ndarray,
        mix_weight: float = 1.0
    ) -> np.ndarray:
        """
        *For internal use only*

        The internal process to blend and mix the provided
        `base_input` and `overlay_input`.

        This method should not force uint8 nor [0, 255]
        range by itself as it would be done in the 'blend'
        main method.
        """
        # We do nothing because the mix is done in the
        # general 'process' method that calls this one
        return overlay_input
    
class AlphaBlenderCPU(_BlenderCPU):
    """
    Blend the different inputs by applying the
    normal and most famous alpha blending and the
    most typical used in video editing.

    This blender will use the alpha channel of the 
    overlay input, multiplied by the `blend_strength`
    parameter provided, to use it as the mixer factor
    between the base and the overlay inputs.
    """

    def _blend(
        self,
        base_input: np.ndarray,
        overlay_input: np.ndarray,
        blend_strength: float = 1.0
    ) -> np.ndarray:
        """
        *For internal use only*

        The internal process to blend and mix the provided
        `base_input` and `overlay_input`.

        This method should not force uint8 nor [0, 255]
        range by itself as it would be done in the 'blend'
        main method.
        """
        base = base_input.astype(np.float32) / 255.0
        overlay = overlay_input.astype(np.float32) / 255.0

        base_rgb, base_a = base[..., :3], base[..., 3:4]
        overlay_rgb, overlay_a = overlay[..., :3], overlay[..., 3:4]

        # Compound alpha
        alpha = overlay_a * blend_strength

        # Equivalent to mix(base, overlay, alpha) in glsl
        out_rgb = base_rgb * (1.0 - alpha) + overlay_rgb * alpha
        out_a = base_a * (1.0 - alpha) + overlay_a * alpha

        result = np.concatenate(
            [out_rgb, out_a],
            axis = -1
        )

        return np.clip(result * 255.0, 0, 255).astype(np.uint8)
    
    def process(
        self,
        base_input: Union[np.ndarray, 'moderngl.Texture'],
        overlay_input: Union[np.ndarray, 'moderngl.Texture'],
        mix_weight: float = 1.0,
        dtype: Union[np.dtype, None] = None,
        blend_strength: float = 0.5
    # TODO: Is this the expected type (?)
    # TODO: What about OpenGL textures (?)
    ) -> np.ndarray:
        return super().process(
            base_input = base_input,
            overlay_input = overlay_input,
            mix_weight = mix_weight,
            dtype = dtype,
            blend_strength = blend_strength
        )
    
class AddBlenderCPU(_BlenderCPU):
    """
    Blend the different inputs by applying the
    normal and most famous alpha blending and the
    most typical used in video editing.

    This blender will increase the brightness by
    combining the colors of the base and the overlay
    inputs, using the overlay as much as the 
    `stregth` parameter is indicating.
    """

    def _blend(
        self,
        base_input: np.ndarray,
        overlay_input: np.ndarray,
        strength: float = 1.0
    ) -> np.ndarray:
        """
        *For internal use only*

        The internal process to blend and mix the provided
        `base_input` and `overlay_input`.

        This method should not force uint8 nor [0, 255]
        range by itself as it would be done in the 'blend'
        main method.
        """
        base = base_input.astype(np.float32) / 255.0
        overlay = overlay_input.astype(np.float32) / 255.0

        # Add blending (base + overlay * strength)
        result = base + overlay * strength

        result = np.clip(result, 0.0, 1.0)

        return np.clip(result * 255.0, 0, 255).astype(np.uint8)
    
    def process(
        self,
        base_input: Union[np.ndarray, 'moderngl.Texture'],
        overlay_input: Union[np.ndarray, 'moderngl.Texture'],
        mix_weight: float = 1.0,
        dtype: Union[np.dtype, None] = None,
        strength: float = 1.0
    # TODO: Is this the expected type (?)
    # TODO: What about OpenGL textures (?)
    ) -> np.ndarray:
        return super().process(
            base_input = base_input,
            overlay_input = overlay_input,
            mix_weight = mix_weight,
            dtype = dtype,
            strength = strength
        )


"""
The nodes that modify inputs by using static parameters
and not 't' time moments, with CPU.
"""
from yta_programming.singleton import SingletonABCMeta
from abc import abstractmethod

import numpy as np


class _NodeProcessorCPU(metaclass = SingletonABCMeta):
    """
    *Singleton class*

    *For internal use only*

    Class to represent a node processor that uses CPU
    to transform the input.

    This class must be implemented by any processor
    that uses CPU to modify an input.
    """
    
    # TODO: Just code and the same attributes that the
    # GPU version also has
    @abstractmethod
    def process(
        self,
        # TODO: What about the type (?)
        input: np.ndarray,
        **kwargs
    # TODO: What about the output type (?)
    ) -> np.ndarray:
        """
        Process the provided 'input' and transform it by
        using the code that is defined here.
        """
        # TODO: Specific attributes can be received as
        # **kwargs to modify the specific process
        pass

# Specific implementations below
class SelectionMaskProcessorCPU(_NodeProcessorCPU):
    """
    Class to use a mask selection (from which we will
    determine if the pixel must be applied or not) to
    apply the `processed_input` on the `original_input`.
    """

    def process(
        self,
        # TODO: What about the type (?)
        original_input: np.ndarray,
        processed_input: np.ndarray,
        selection_mask_input: np.ndarray
    ):
        """
        Apply the `selection_mask` provided to the also
        given `original` and `processed` nuumpy arrays to
        obtain the processed one but affected only as the
        selection mask says.

        The selection mask must have values in the range
        [0.0, 1.0].
        """
        # Float range and normalize if needed
        original_input = original_input.astype(np.float32) / 255.0
        processed_input = processed_input.astype(np.float32) / 255.0

        # We need a 3D or 4D mask
        if selection_mask_input.ndim == 2:
            selection_mask_input = np.expand_dims(selection_mask_input, axis = -1)
        if (
            selection_mask_input.shape[-1] == 1 and
            original_input.shape[-1] in (3, 4)
        ):
            selection_mask_input = np.repeat(selection_mask_input, original_input.shape[-1], axis = -1)

        # Mix with the selection mask
        final = original_input * (1.0 - selection_mask_input) + processed_input * selection_mask_input

        # Back to [0, 255]
        return np.clip(
            a = final * 255.0,
            a_min = 0,
            a_max = 255
        ).astype(np.uint8)


"""
TODO: This module doesn't use 't' but 'progress'
so it is not a child of 'processor.video', maybe
we should move it to be 'processor.transitions'
instead of 'processor.video.transitions'... (?)
"""
# TODO: Should we keep this functionality (?)
# from yta_video_opengl.utils import texture_to_frame
from yta_programming.singleton import SingletonABCMeta
from yta_validation import PythonValidator
from abc import abstractmethod
from typing import Union

import numpy as np


class _TransitionProcessorCPU(metaclass = SingletonABCMeta):
    """
    *Abstract class*

    *For internal use only*

    A transition between the frames of 2 videos.

    This transition is made with CPU (numpy).
    """

    @abstractmethod
    def process(
        self,
        # TODO: What about the type (?)
        first_input: Union['moderngl.Texture', np.ndarray],
        second_input: Union['moderngl.Texture', np.ndarray],
        progress: float,
        **kwargs
    ) -> np.ndarray:
        """
        Process the `first_input` and `second_input` and
        generate the transition frame according to the
        `progress` of the transition provided.
        """
        # TODO: Specific attributes can be received as
        # **kwargs to modify the specific process
        pass

class SlideTransitionProcessorCPU(_TransitionProcessorCPU):
    """
    A transition in which the frames goes from one
    side to the other, disappearing the first one
    and appearing the second one.
    """

    def process(
        self,
        # TODO: What about the type (?)
        first_input: Union['moderngl.Texture', np.ndarray],
        second_input: Union['moderngl.Texture', np.ndarray],
        progress: float
    ) -> np.ndarray:
        """
        Process the `first_input` and `second_input` and
        generate the transition frame according to the
        `progress` of the transition provided.
        """
        # TODO: Should we keep this functionality (?)
        # first_input = (
        #     texture_to_frame(first_input, do_include_alpha = True)
        #     if PythonValidator.is_instance_of(first_input, 'moderngl.Texture') else
        #     first_input
        # )

        # second_input = (
        #     texture_to_frame(second_input, do_include_alpha = True)
        #     if PythonValidator.is_instance_of(second_input, 'moderngl.Texture') else
        #     second_input
        # )

        # TODO: What do we do in this case (?)
        assert first_input.shape == second_input.shape, "Los frames deben tener la misma forma"

        h, w, c = first_input.shape
        frame_out = np.zeros_like(first_input)

        # By default we are only handling 'left' direction
        direction = 'left'

        offset_x = 0
        # offset_y = 0
        if direction == 'left':
            offset_x = int(w * progress)
            if offset_x < w:
                frame_out[:, :w - offset_x] = first_input[:, offset_x:]
            if offset_x > 0:
                frame_out[:, w - offset_x:] = second_input[:, :offset_x]

        # elif direction == 'right':
        #     offset_x = int(w * progress)
        #     if offset_x < w:
        #         frame_out[:, offset_x:] = first_input[:, :w - offset_x]
        #     if offset_x > 0:
        #         frame_out[:, :offset_x] = second_input[:, w - offset_x:]

        # elif direction == 'up':
        #     offset_y = int(h * progress)
        #     if offset_y < h:
        #         frame_out[:h - offset_y, :] = first_input[offset_y:, :]
        #     if offset_y > 0:
        #         frame_out[h - offset_y:, :] = second_input[:offset_y, :]

        # elif direction == 'down':
        #     offset_y = int(h * progress)
        #     if offset_y < h:
        #         frame_out[offset_y:, :] = first_input[:h - offset_y, :]
        #     if offset_y > 0:
        #         frame_out[:offset_y, :] = second_input[h - offset_y:, :]

        return frame_out